functor ContextFun(structure Throttle' : THROTTLE
		   structure Dataset' : DATASET
		   structure AcapTime' : ACAPTIME
		   sharing AcapTime' = Dataset'.AcapTime
		   sharing type Dataset'.T.throttle = Throttle'.throttle)
  :> CONTEXT 
  where type Dataset.export = Dataset'.export 
    and type Dataset.updatemsg = Dataset'.updatemsg
    and type Dataset.Acl.acl = Dataset'.Acl.acl
    and type Dataset.Acl.Auth.ident = Dataset'.Acl.Auth.ident
    and type Dataset.Acl.Rights.rights = Dataset'.Acl.Rights.rights
    and type 'a Throttle.throttle = 'a Throttle'.throttle 
    and type AcapTime.acaptime = AcapTime'.acaptime =
struct
  structure Throttle = Throttle'
  structure Dataset = Dataset'
  structure AcapTime = AcapTime'
  structure E = Dataset
  structure Auth = Dataset.Acl.Auth

  (* messages to contexts *)
  datatype inMsg = 
    ADD of (string * Dataset.export)
  | DONEADD
  | ADDEVENT of (string * Dataset.updatemsg CML.event)
  | DELETECONTEXT
  | UPDATECONTEXT
  | SEARCH of ((int option * Dataset.export) -> bool) * 
               (string * Dataset.export) Throttle.throttle *
	       (AcapTime.acaptime * (AcapTime.acaptime -> unit))

  datatype outMsg = 
    CNEW of Dataset.export * int option
  | CCHANGE of Dataset.export * (int * int) option
  | CDELETE of Dataset.export * int option
  | CTIME of AcapTime.acaptime
  | CDIE

  type search = (int option * E.export) -> bool
       (* int option is the place in the context *)
  type sort = ((string * E.export) * (string * E.export)) -> order

  type condata = (string * E.export) list

  type context = {guard : E.export -> bool,
		  notify : outMsg CML.chan option,
		  owner : Auth.ident,
		  sort : sort option,
		  enum : bool,
		  output : condata -> unit,
		  delete : condata -> (string * E.export) -> condata,
		  change : condata -> (string * (E.export * E.export))
		     -> condata,
		  insert : condata -> (string * E.export) -> condata,
		  time : AcapTime.acaptime}

  (* search string
     messenger option NOTIFY
     sort option SORT
     bool ENUMERATE *)
  (* val create : (E.entry -> bool) -> msg CML.channel option -> sort option 
               -> bool -> E.entry Throttle.throttle -> msg CML.channel *)

  exception NotSupported
  exception NotImplemented
  exception BadEntry

  (* empty context *)
  val empty = nil
    
  fun doInsert owner NONE NONE data e = e::data
    | doInsert owner (SOME c) NONE data (p, e) = 
    ((CML.send (c, CNEW (e, NONE))); 
     (p,e)::data)
    | doInsert owner notify (SOME sort) data (p, e) =
    let
      val notify = 
	case notify of 
	  NONE => (fn (n) => ())
	| (SOME c) => (fn n => (CML.send (c, CNEW(e, SOME n))))

      fun ins n (nil) = (notify n; (p,e)::nil)
	| ins n (a::l) = 
	case sort (a, (p, e)) of
	  LESS => a::(ins (n+1) l)
	| EQUAL => (notify (n+1); a::(p,e)::l)
	| GREATER => (notify (n); (p,e)::a::l)
    in
      ins 1 data
    end

  fun doChange owner notify NONE data (p, (old, new)) =
    let
      val name = E.nameEntry (owner, old)

      val _ = 
	case notify of 
	  NONE => ()
	| (SOME c) =>
	  if name = E.nameEntry (owner, new) then
	    CML.send (c, CCHANGE(new, NONE))
	  else (CML.send (c, CDELETE(old, NONE));
		CML.send (c, CNEW(new, NONE)))

      fun chg (nil) = raise BadEntry
	| chg ((p', a)::l) = 
	if (p = p') andalso (name = E.nameEntry (owner, a)) 
	  then (p, new)::l
	else (p', a)::(chg l)
    in
      chg data
    end
    | doChange owner notify (SOME sort) data (p, (old, new)) =
    let
      val name = E.nameEntry (owner, old)
      val newname = E.nameEntry (owner, new)

      val notify = 
	case notify of 
	  NONE => (fn (pos) => ())
	| (SOME c) =>
	  if name = newname then
	    (fn pos => CML.send (c, CCHANGE(new, SOME pos)))
	  else (fn (p1, p2) => (CML.send (c, CDELETE(old, SOME p1));
				CML.send (c, CNEW(new, SOME p2))))

      val oldpos = ref 0
      val newpos = ref 0

      fun chg n (nil) = raise BadEntry
	| chg n ((p', a)::l) = 
	case sort ((p',a), (p,old)) of
	  LESS => (p', a)::(chg (n+1) l)
	| EQUAL => 
	    if (p = p') andalso (name = E.nameEntry (owner, a)) 
	      then (oldpos := n; l)
	    else (p', a)::(chg (n+1) l)
	| GREATER => raise BadEntry

      val data = chg 1 data
	
      fun ins n (nil) = (newpos := n; (p,new)::nil)
	| ins n (a::l) = 
	case sort (a, (p,new)) of
	  LESS => a::(ins (n+1) l)
	| EQUAL => (newpos := n+1; a::(p,new)::l)
	| GREATER => (newpos := n; (p,new)::a::l)

      val data = ins 1 data

      val _ = notify(!oldpos, !newpos)
    in
      data
    end

  fun doDelete owner notify _ data (p, e) =
    let
      val name = E.nameEntry (owner, e)
      fun del n (nil) = raise BadEntry
	| del n ((p', a)::l) = 
	let
	  val notify =
	    case notify of
	      NONE => (fn n => ())
	    | (SOME c) => (fn n => (CML.send (c, CDELETE(e, SOME n))))
	in
	  if (p = p' andalso name = E.nameEntry (owner, a)) then (notify n; l)
	  else (p', a)::(del (n+1) l)
	end
    in
      del 1 data
    end


  fun doOutput out nil = Throttle.done out
    | doOutput out ((p, a)::l) = (Throttle.call out (p, a); doOutput out l)

  fun doSearch out search data = 
    (let
       fun doit nil n = Throttle.done out
	 | doit ((p, a)::l) n =
	 (if (search (SOME(n), a)) then Throttle.call out (p, a) else ();
	  doit l (n+1))
     in
       doit data 1
     end handle AbortSearch => ())

  fun doUpdate (context as {guard, notify, owner, sort,
			    insert, enum, time,
			    output, delete, 
			    change}, data, evts) (p, E.NEW (e)) =
    if guard e then ({guard=guard, notify=notify, owner=owner, sort=sort,
			    insert=insert, enum=enum,
			    output=output, delete=delete, 
			    change=change, time=AcapTime.now ()}, 
		     (insert data (p,e)), evts)
    else (context, data, evts)
    | doUpdate (context as {guard, notify, owner, sort, 
			    insert, enum, time,
			    output, delete, 
			    change}, data, evts) (p, E.CHANGE (old, new)) =
      if guard old then 
	if guard new then ({guard=guard, notify=notify, owner=owner,
			    sort=sort,
			    insert=insert, enum=enum,
			    output=output, delete=delete, 
			    change=change, time=AcapTime.now ()}, 
			   (change data (p, (old, new))), evts)
	else (context, (delete data (p, old)), evts)
      else (context, (insert data (p, new)), evts)
   | doUpdate (context as {guard, notify, owner, sort,
			   insert, enum, time,
			   output, delete, 
			   change}, data, evts) (p, E.DELETE e) =
	if guard e then ({guard=guard, notify=notify, owner=owner, sort=sort,
			    insert=insert, enum=enum,
			    output=output, delete=delete, 
			    change=change, time=AcapTime.now ()}, 
			 (delete data (p, e)), evts)
	else (context, data, evts)

  datatype inmsgs = 
    DS of string * E.updatemsg
  | CMD of inMsg

  fun doCmdStatic (context as {guard, notify, owner, 
			       insert, enum, sort, time,
			       output, delete, change}, data, evts) evt =
    case evt of
      (ADD e) => (context, (insert data e), evts)
    | (DONEADD) => (output data; (context, data, evts))
    | (ADDEVENT evt) => (context, data, evts)
    | DELETECONTEXT => CML.exit ()
    | UPDATECONTEXT => raise NotSupported
    | (SEARCH (search, out, (t, cancelTime))) => 
	(if AcapTime.compare (time, t) = GREATER then
	   cancelTime time
	 else doSearch out search data;
	 (context, data, evts))

  fun DS' p = (fn x => DS(p, x))

  fun doCmdDyn (context as {guard, notify=(SOME c), owner,
			    insert, enum, sort, time,
			    output, delete, change} : context, 
		data, evts) evt =
    case evt of	 
      (ADD (p, e)) => ({guard=guard,
			notify=(SOME c),
			owner=owner,
			enum=enum,
			output=output,
			sort=sort,
			insert=insert,
			delete=delete,
			change=change,
			time=E.timeEntry e}, (insert data (p, e)), evts)
    | (DONEADD) => (* now enable new, change, etc. *)
	(output data;
	 ({guard=guard, notify=(SOME c), owner=owner,
	   enum=enum, output=output, sort=sort,
	   time=AcapTime.now (),
	   insert=doInsert owner (SOME c) sort,
	   delete=doDelete owner (SOME c) sort,
	   change=doChange owner (SOME c) sort},
	  data, evts))
    | (ADDEVENT (p,evt)) => (context, data, ((CML.wrap (evt, DS' p))::evts))
    | DELETECONTEXT => (CML.send (c, CDIE); CML.exit ())
    | UPDATECONTEXT => (* send time notification *)
	(CML.send (c, CTIME time);
	 (context, data, evts))
    | (SEARCH (search, out, (t, cancelTime))) => 
	(if AcapTime.compare (time, t) = GREATER then
	   cancelTime time
	 else doSearch out search data;
	 (context, data, evts))
	    
  fun listen (p as ({guard, notify = NONE, owner, 
		     time, insert, enum, sort,
		     output, delete, change} : context, data, evts)) = 
    (* static context; no notifications supported! *)
    (case CML.select evts of
       (CMD c) => (print "got CMD!\n"; listen (doCmdStatic p c))
     | (DS d) => (print "got DS!\n"; raise NotSupported))
    | listen (p as ({guard, notify = SOME _, owner, 
		     time, insert, enum, sort,
		    output, delete, change}, data, evts)) = 
       (* dynamic context; we need to delete with NEWs, etc. *)
       (case CML.select evts of
	  (CMD c) => listen (doCmdDyn p c)
	| (DS d) => listen (doUpdate p d))

  fun create ident guard notify (sort : sort option) enum out =
       let
	 val ch = CML.channel ()
	 val evt = CML.wrap (CML.recvEvt ch, CMD)
	 val _ = print "creating thread!\n"
	 val th = CML.spawn (fn () => 
			     (listen ({guard=guard, 
				       notify=notify, 
				       owner=ident,
				       insert=doInsert ident NONE sort,
				       change=doChange ident NONE sort,
				       delete=doDelete ident NONE sort,
				       sort=sort,
				       enum=enum, 
				       time = AcapTime.now (),
				       output=doOutput out} : context, 
				      empty, [evt])))
       in
	 ch
       end

end
