(* $Id: dataset.sig,v 1.11 1999/10/11 20:57:53 leg Exp $ *)

signature DATASET =
sig
  structure T : THROTTLE
  structure AcapTime : ACAPTIME
  structure Acl : ACL

  type dataset
  type export
  type metadata = {acl : Acl.acl option, attribute : string,
		   size : int list, value : Value.value}
  type storedata = {acl : Acl.acl option, attribute : string, 
		    value : Value.value option}
  type attribute = metadata

  (* --- entries --- *)
  (* raised when an attempt to store an illegal value is made to this
   dataset---returns entry name/attribute raised on *)
  exception EnforcedValidation of (string * string)

  (* raised when an attempt is made to store with a UNCHANGEDSINCE;
     returns the time it was last modified *)
  exception Modified of AcapTime.acaptime

  (* syntax error detected during store *)
  exception SyntaxError of string

  (* rights violation *)
  exception NotPermitted

  (* gets the attributes user * doinh * entry -> attrname -> attr *)
  val fetchEntry : (Acl.Auth.ident * bool * export) -> string -> attribute

  (* gets the matching attributes 
   user * doinh * entry -> searchfun -> attrlist *)
  val searchEntry : (Acl.Auth.ident * bool * export) -> (string -> bool) 
                                                     -> attribute list

  val timeEntry : export -> AcapTime.acaptime

  (* get the name *)
  val nameEntry : (Acl.Auth.ident * export) -> string

  (* --- datasets --- *)
  exception NoSuchDataset
  exception AlreadyExists of dataset
  exception DatasetRefer of string list
  exception RenameEntry of string

  datatype updatemsg =
    NEW of export
  | CHANGE of (export * export)
  | DELETE of export

  val init : unit -> unit

  (* search the named dataset 
   params: the user,
           the named dataset,
	   the depth,
           the search function, 
	   a messenger to leave at each dataset to be called,
	   the throttle to send output to,
	   and doinh (true -> yes, inherit)
	   output comes in (path : string, entry) form. *)
  val search : Acl.Auth.ident -> string list -> int
    -> (export -> bool)
    -> (string * updatemsg Multicast.port -> unit) option 
    -> (string * export) T.throttle -> bool -> unit

  (* store the list of entry/attribute pairs in this dataset;
   entries should be _full_ name
   params: the user,
           the dataset path
           whether to create or not
           what to store
           time to abort *)
  val store : Acl.Auth.ident -> string list -> bool 
    -> (string * storedata list) list 
    -> AcapTime.acaptime option -> unit

  (* acl stuff *)
  type aclobj = {dataset : string list,
		 attribute : string option,
		 entry : string option}

  datatype aclmod = 
    SETACL of string * Acl.Rights.rights
  | DELACL of string option
  | GETACL

  val modifyacl : Acl.Auth.ident -> (aclobj * aclmod) -> Acl.acl

end
