functor ExecuteFun(structure Acl' : ACL
		   structure Dataset : DATASET
		   structure Throttle : THROTTLE
		   structure Context : CONTEXT
		   structure Comparator : COMPARATOR
		   structure AcapTime : ACAPTIME
	sharing Context.Dataset.Acl = Dataset.Acl = Acl'
	sharing Dataset.AcapTime = Context.AcapTime = AcapTime
	sharing type Context.Dataset.export = Dataset.export
        sharing type Context.Dataset.updatemsg = Dataset.updatemsg
	sharing type Context.Dataset.Acl.Auth.ident = Dataset.Acl.Auth.ident
        sharing type Throttle.throttle = Context.Throttle.throttle 
		     = Dataset.T.throttle)
  :> EXECUTE =
struct
  structure A = Absyn
  structure R = Response

  structure M = Mailbox

  structure Acl = Acl'

  exception CloseConnection
  exception doAuth of Sasl.conn * string * string * string option 
                                * R.response Mailbox.mbox

  (* internal error *)
  type conn = {user : Acl.Auth.ident option, 
	       comm : R.response M.mbox,
	       contexts : (string * bool * bool * Context.inMsg CML.chan) list,
	       saslconn : Sasl.conn}

  fun start comm sasl = ({user=NONE, comm=comm, contexts=nil,
			  saslconn=sasl} : conn)

  fun authAs ({user, comm, contexts, saslconn} : conn) u =
    (print "authenticating...\n";
     {user=SOME u, comm=comm, 
      contexts=contexts, saslconn=saslconn} : conn)

  exception BadSearchCriteria
  exception BadSearchModifier
  exception BadAttributePattern
  exception BadAttributeMetadata
  exception NoSuchContext
  exception BadEntryPath
  exception BadStoreModifier
  exception BadStoreMetadata

  exception AbortSearch

  exception InternalError of string

  fun splitDSet auth d =
    let
      val path = 
	case String.fields (fn c => c = #"/") d of
	  nil => raise BadEntryPath
	| (""::path) => path
	| _ => raise BadEntryPath

      fun doSub s = if s = "~" then (Acl.Auth.toString auth) else s
    in
      List.map doSub path
    end

  fun deleteContext contexts c =
    let
      val (tofree, others) = List.partition (fn (n,_,_,_) => n = c) contexts
    in
      case tofree of
	nil => raise NoSuchContext
      | [(_, _, _, chan)] => (CML.send (chan, Context.DELETECONTEXT);
			      others)
      | _ => raise (InternalError "weird context list")
    end

  fun compileDSearch (ident, doinh) spec =
    let
      val myfetch = fn e => Dataset.fetchEntry (ident, doinh, e)

      fun compileDSearch' A.All = (fn e => true)
	| compileDSearch' (A.Equal (attr, (A.Comparator comp), v)) =
	let
	  val comp = Comparator.lookup Comparator.Equal comp
	  fun f e = (case myfetch e attr of
		       {value,...} => (comp(v, value) = EQUAL))
	in f end
	| compileDSearch' (A.Compare (attr, (A.Comparator comp), v)) =
	let
	  val comp = Comparator.lookup Comparator.Order comp
	  fun f e = (case myfetch e attr of
		       {value,...} => (comp(v, value) <> GREATER))
	in f end
	| compileDSearch' (A.Prefix (attr, (A.Comparator comp), v)) =
	let
	  val comp = Comparator.lookup Comparator.Prefix comp
	  fun f e = (case myfetch e attr of
		       {value,...} => (comp(v, value) = EQUAL))
	in f end
	| compileDSearch' (A.Range _) = raise BadSearchCriteria
	| compileDSearch' (A.Strict (attr, (A.Comparator comp), v)) =
	let
	  val comp = Comparator.lookup Comparator.Order comp
	  fun f e = (case myfetch e attr of
		       {value,...} => (comp(v, value) = LESS))
	in f end
	| compileDSearch' (A.Substr (attr, (A.Comparator comp), v)) =
	let
	  val comp = Comparator.lookup Comparator.Substring comp
	  fun f e = (case myfetch e attr of
		       {value,...} => (comp(v, value) = EQUAL))
	in f end
	| compileDSearch' (A.Not crit) =
	let
	  val f = compileDSearch' crit
	in not o f end
	| compileDSearch' (A.Or (crit1, crit2)) =
	let
	  val f = compileDSearch' crit1
	  val g = compileDSearch' crit2
	in
	  (fn e => f e orelse g e)
	end
	| compileDSearch' (A.And (crit1, crit2)) =
	let
	  val f = compileDSearch' crit1
	  val g = compileDSearch' crit2
	in
	  (fn e => f e andalso g e)
	end
    in
      compileDSearch' spec
    end

  (* returns a search fun and a time that the search is good up to *)
  fun compileCSearch (ident, doinh) spec =
    let
      val myfetch = fn e => Dataset.fetchEntry (ident, doinh, e)

      fun compileCSearch' enum A.All = (fn e => true, AcapTime.future)
	| compileCSearch' enum (A.Equal (attr, (A.Comparator comp), v)) =
	let
	  val comp = Comparator.lookup Comparator.Equal comp
	  fun f (_,e) = (case myfetch e attr of
			   {value,...} => (comp(v, value) = EQUAL))
	in (f, AcapTime.future) end
	| compileCSearch' enum (A.Compare (attr, (A.Comparator comp), v)) =
	let
	  val comp = Comparator.lookup Comparator.Order comp
	  fun f (_,e) = (case myfetch e attr of
			   {value,...} => (comp(v, value) <> GREATER))
	in (f, AcapTime.future) end
	| compileCSearch' enum (A.Prefix (attr, (A.Comparator comp), v)) =
	let
	  val comp = Comparator.lookup Comparator.Prefix comp
	  fun f (_,e) = (case myfetch e attr of
			   {value,...} => (comp(v, value) = EQUAL))
	in (f, AcapTime.future) end
	| compileCSearch' enum (A.Range (s, e, t)) = 
	  if not enum then raise BadSearchCriteria
	  else let
		 fun f (SOME i, _) = s <= i andalso i <= e
		   | f (NONE, _) = raise 
		   (InternalError "enumerated context but no enumeration?!?")
	       in
		 (f, AcapTime.fromString t)
	       end
	| compileCSearch' enum (A.Strict (attr, (A.Comparator comp), v)) =
	let
	  val comp = Comparator.lookup Comparator.Order comp
	  fun f (_,e) = (case myfetch e attr of
			   {value,...} => (comp(v, value) = LESS))
	in (f, AcapTime.future) end
	| compileCSearch' enum (A.Substr (attr, (A.Comparator comp), v)) =
	let
	  val comp = Comparator.lookup Comparator.Substring comp
	  fun f (_,e) = (case myfetch e attr of
			   {value,...} => (comp(v, value) = EQUAL))
	in (f, AcapTime.future) end
	| compileCSearch' enum (A.Not crit) =
	let
	  val (f, t) = compileCSearch' enum crit
	in (not o f, t) end
	| compileCSearch' enum (A.Or (crit1, crit2)) =
        let
	  val (f, t1) = compileCSearch' enum crit1
	  val (g, t2) = compileCSearch' enum crit2
	  val t = 
	    (case AcapTime.compare (t1, t2) of
	       GREATER => t2
	     | LESS => t1
	     | EQUAL => t1)
	in
	  ((fn e => f e orelse g e), t)
	end
	| compileCSearch' enum (A.And (crit1, crit2)) =
	let
	  val (f, t1) = compileCSearch' enum crit1
	  val (g, t2) = compileCSearch' enum crit2
	  val t = 
	    (case AcapTime.compare (t1, t2) of
	       GREATER => t2
	     | LESS => t1
	     | EQUAL => t1)
	in
	  ((fn e => f e andalso g e), t)
	end
    in
      compileCSearch' spec
    end

  fun compileSort (ident, doinh) sortlist =
    let
      val myfetch = fn e => Dataset.fetchEntry (ident, doinh, e)

      fun compileSort' nil = (fn (e1, e2) => EQUAL)
	| compileSort' (A.SortItem(attr, A.Comparator comp)::l) =
	let
	  val comp = Comparator.lookup Comparator.Order comp
	  val f = compileSort' l
	  fun getV (_, e) = #value (myfetch e attr)
	  fun g (e1, e2) = case comp (getV e1, getV e2) of
	    LESS => LESS
	  | EQUAL => f (e1, e2)
	  | GREATER => GREATER
	in
	  g
	end
    in
      compileSort' sortlist
    end
      
  (* returns a function to apply to an entry;
   that functions returns an entry suitable for appending to an ENTRY
   list *)
  fun compileMetaData (ident, doinh) nil = (fn e => "")
    | compileMetaData (ident, doinh) metadata_list = 
      let
	val myfetch = fn e => Dataset.fetchEntry (ident, doinh, e)
	val mysearch = fn e => Dataset.searchEntry (ident, doinh, e)
	
	(* we have to figure out what attributes to pull; we use
	 search for patterns, fetch otherwise *)
	datatype tomatch = pat of string | lit of string

	(* should probably verify this is an ok attribute o/w *)
	fun isPat nil s = lit (String.implode (List.rev s))
	  | isPat [#"*"] s = pat (String.implode (List.rev s))
	  | isPat [c] s = lit (String.implode (List.rev (c::s)))
	  | isPat (#"*"::s) s2 = raise BadAttributePattern
	  | isPat (c::s) s2 = isPat s (c::s2)
	  
        (* construct a function to extract the appropriate metadata out *)
	fun getmeta nil = (fn _ => nil)
	  | getmeta ("attribute"::l) =
	  let val rest = getmeta l in
	    (fn (md as {attribute,...}:Dataset.attribute) =>
	     ("\"" ^ attribute ^ "\"")::(rest md))
	  end
	  | getmeta ("value"::l) =
	  let val rest = getmeta l in
	    (fn (md as {value,...}:Dataset.attribute) =>
	     (Value.toString value)::(rest md))
	  end
	  | getmeta ("acl"::l) =
	  let val rest = getmeta l in
	    (fn (md as {acl,...}:Dataset.attribute) =>
	     (Acl.optToString acl)::(rest md))
	  end
	  | getmeta ("myrights"::l) =
	  let val rest = getmeta l in
	    (fn (md) => "yuck"::(rest md))
	  end
	  | getmeta ("size"::l) = 
	  let val rest = getmeta l in
	    (fn (md as {size,...}:Dataset.attribute) =>
	     (map Int.toString size) @ (rest md))
	  end
	  | getmeta (_::l) = raise BadAttributeMetadata
	  
	(* we must have at least one metadata item ! *)
	(* formatMeta : string list -> string *)
	fun formatMeta nil = raise (InternalError "need one metadata item")
	  | formatMeta [s] = s
	  | formatMeta (s::l) = "(" ^ s ^ 
	        (List.foldr (fn (s, acc) => (" " ^ s ^ acc)) ")" l)
	  
        (* this function returns a list of functions we apply to the entry.
	 we concat them all together to get the return-data-list.

	 this would be more efficient if we could do one fetch for all
	 the static data items; right now that's too hard and i punt *)
	fun fetchit nil fl = List.rev fl
	  | fetchit (A.MetaData(a, md)::l) fl = 
	  (case isPat (String.explode a) nil of
	     (pat s) => let 
	                  (* what attributes to print out *)
			  val match = String.isPrefix s 
			    
			  (* what metadata do we want from these attributes? *)
			  val print = if List.null md then 
			                ["attribute", "value"]
				      else md
					
			  (* val sl : attribute -> string *)
			  val sl = formatMeta o (getmeta print)

                          (* Entry.search e match will give us a list of
			   attributes. *)

		    (* we map sl over it to get a list of return-metalist,
		     which we then put together to form a return-attr-list,
		     which is part of the return-data-list *)
			  val thisFn = 
			  (fn e => case List.map sl (mysearch e match) of
			             nil => "()"
				   | [a] => "(" ^ a ^ ")"
				   | (a::l) => "(" ^ a ^ 
		       (List.foldr (fn (s,acc) => (" " ^ s ^ acc)) "" l) ^ ")")
			in fetchit l (thisFn::fl)
			end
	   | (lit s) => let
			(* hopefully easier than above; requested a single
			 attribute *)

			  (* default metadata *)
			  val print = if List.null md then ["value"]
				      else md

			  (* val sl : attribute -> string *)
			  val sl = formatMeta o (getmeta print) 

                          val thisFn = 
			    (fn e => sl (myfetch e s))
			in fetchit l (thisFn::fl)
			end)

	(* ok, for those of you not paying attention, we now apply
	 fetchit to our metadata list to give us a list of functions
	 we must apply to the entry; when an entry is returned, we apply
	 those functions and smoosh it all together to form a
	 return-data-list, the last parameter to R.ENTRY *)
	val fl = fetchit metadata_list nil

	fun makeString e =
	  let
	    val retdatalist = map (fn f => f e) fl
	    val retdatalist = 
	      case retdatalist of
		nil => ""
	      | [s] => s
	      | (s::l) => s ^ (List.foldr (fn (s,acc) => (" " ^ s ^ acc)) "" l)
	  in
	    retdatalist
	  end
      in
	makeString
      end

  fun compileRet _ entname NONE = (fn (p, e) => ()) (* throw 'em away! *)
    | compileRet (comm, tag) entname (SOME rdatafun) = 
      (fn (p, e) => (M.send (comm, R.ENTRY(tag, entname (p, e), rdatafun e))))

  (* this is how we return stuff back *)
  fun donePass (tag, comm) =
    let
      val retmodtime = R.TAGMODTIME(tag,
				    (AcapTime.toString o AcapTime.now) ())
	
      fun donePass (Throttle.OK) = 
	(M.send (comm, retmodtime);
	 M.send (comm, R.OK(SOME tag, nil, "search completed")))
    in
      donePass
    end

  fun doneSoft (tag, comm) =
    let
      val retmodtime = R.TAGMODTIME(tag,
				    (AcapTime.toString o AcapTime.now) ())
	
      fun doneSoft (Throttle.S_OK) = 
	(M.send (comm, retmodtime);
	 M.send (comm, R.OK(SOME tag, nil, "search completed")))
	| doneSoft (Throttle.S_TOOMANY i) =
	(M.send (comm, retmodtime);
	 M.send (comm, R.OK(SOME tag, [R.TOOMANY i], "search completed")))
    in
      doneSoft
    end


  fun searchdset (conn as {comm,contexts,user=(SOME user),saslconn=saslconn},
		  tag, dset, sm, sc) =
    (let
       (* drop the first empty field *)
       val dsetpath = splitDSet user dset

       (* extract options *)
       fun get nil (opts as (depth, context, hard, soft, ret, sort, doinh)) = 
	                   opts
	 | get ((A.Depth x)::l) (NONE,c,h,s,r,t,i) = 
			   get l (SOME x, c,h,s,r,t,i)
	 | get ((A.HardLimit x)::l) (d,c,NONE,s,r,t,i) = 
			   get l (d,c,SOME x,s,r,t,i)
	 | get ((A.Limit x)::l) (d,c,h,NONE,r,t,i) = 
			   get l (d,c,h,SOME x,r,t,i)
	 | get ((A.MakeContext x)::l) (d,NONE,h,s,r,t,i) = 
			   get l (d,SOME x,h,s,r,t,i)
	 | get ((A.Return x)::l) (d,c,h,s,NONE,t,i) = 
			   get l (d,c,h,s,SOME x,t,i)
	 | get ((A.Sort x)::l) (d,c,h,s,r,NONE,i) = 
			   get l (d,c,h,s,r,SOME x,i)
	 | get ((A.Noinherit)::l) (d,c,h,s,r,t,NONE) =
			   get l (d,c,h,s,r,t, SOME false)
						     
	 | get _ _ = raise BadSearchModifier

       val (depth, context, hard, soft, ret, sort, doinh) = 
	     get sm (NONE, NONE, NONE, NONE, NONE, NONE, NONE)

       val doinh = getOpt (doinh, true)

       (* compile the search criteria *)
       val search = compileDSearch (user, doinh) sc

       (* depth defaults to 1 *)
       val depth = getOpt (depth, 1)

       (* compile sort criteria *)
       val sort = Option.map (compileSort (user, doinh)) sort

       (* compile the return attributes; yields a function that sends to
	the client *)
       val metaret = Option.map (compileMetaData (user, doinh)) ret

       val ret = compileRet (comm, tag) 
	 (if depth <> 1 then 
	    fn (p, e) => (p ^ "/" ^ (Dataset.nameEntry (user, e)))
	  else fn (p, e) => Dataset.nameEntry (user, e)) metaret

       (* compile the soft limit *)
       val softThrottle = 
	 case soft of
	   NONE => Throttle.createpass ret (donePass (tag, comm))
	 | (SOME(a,b)) => Throttle.createsoft (a,b) ret (doneSoft (tag, comm))

       (* compile the context *)
       val (contexts, context) = 
	 case context of
	   NONE => (contexts, NONE) (* just pass it through *)
	 | SOME(enum, false, name) => (* no notifications *)
	     let
	       val contexts = ((deleteContext contexts name)
		 handle NoSuchContext => contexts)

	       val c = 
		 Context.create user search NONE sort enum softThrottle
	     in
	       ((name, enum, false, c)::contexts, SOME c)
	     end
	 | SOME(enum, true, name) => (* notifications *)
	     let
	       val contexts = ((deleteContext contexts name)
		 handle NoSuchContext => contexts)

	       (* create output channel *)
	       val outchan = CML.channel ()

	       val metaret = getOpt(metaret, (fn e => ""))

	       (* function to convert notifications into output *)
	       fun conv (Context.CNEW (e, (SOME i))) =
		 M.send (comm, R.ADDTO(name, Dataset.nameEntry (user, e),
				       i, metaret e))
		 | conv (Context.CNEW (e, NONE)) =
		 M.send (comm, R.ADDTO(name, Dataset.nameEntry (user, e),
				       0, metaret e))
		 | conv (Context.CCHANGE (e, (SOME (i, j)))) =
		 M.send (comm, R.CHANGE(name, Dataset.nameEntry (user, e),
					i, j, metaret e))
		 | conv (Context.CCHANGE (e, NONE)) = 
		 M.send (comm, R.CHANGE(name, Dataset.nameEntry (user, e),
					0, 0, metaret e))
		 | conv (Context.CDELETE (e, (SOME i))) =
		 M.send (comm, R.REMOVEFROM(name, Dataset.nameEntry (user, e),
					    i))
		 | conv (Context.CDELETE (e, NONE)) =
		 M.send (comm, R.REMOVEFROM(name, Dataset.nameEntry (user, e), 
					    0))
		 | conv (Context.CTIME time) =
		 M.send (comm, R.MODTIME(name, AcapTime.toString time))
		 | conv (Context.CDIE) = CML.exit ()
		 
	       (* and a process to convert between the two *)
	       fun lconv ch = (conv (CML.recv ch); lconv ch)
		 
	       val _ = CML.spawnc lconv outchan
		 
	       val c = Context.create user search (SOME outchan) 
                                      sort enum softThrottle
	     in
	       ((name, enum, true, c)::contexts, SOME c)
	     end

       (* compile the sort throttle, if needed *)
       val sortThrottle =
	 (case (sort, context) of
	    (NONE, NONE) => softThrottle (* skip it *)
	  | (SOME f, NONE) => 
	      Throttle.createsort f (Throttle.call softThrottle) 
	                          (fn OK => Throttle.done softThrottle)
	  | (_, SOME c) => (* if there's a context, it'll do the sorting *)
	      Throttle.createpass (fn e => CML.send (c, Context.ADD e))
	                          (fn OK => CML.send (c, Context.DONEADD)))

       fun doneHard f (Throttle.H_OK) = f () (* call the done on the next fn *)
	 | doneHard f (Throttle.H_WAYTOOMANY) =
	 let
	   (* ok, this should kill the search in progress, report failure,
	    and kill the context we're building, if any *)

	   (* send NO *)
	   val _ = M.send (comm, R.NO(SOME tag, [R.WAYTOOMANY],
			              "hard limit exceeded"))

	   (* remove context *)
	   val _ = (case context of SOME c => CML.send (c, Context.DELETECONTEXT)
                           	  | NONE => ())
	 in
	   (* finally, tell the search to stop *)
	   raise AbortSearch
	 end

       (* compile the hard throttle *)
       val hardThrottle = 
	 (case hard of
	    NONE => sortThrottle (* skip it *)
	  | (SOME x) => Throttle.createhard x (Throttle.call sortThrottle) 
	                  (doneHard (fn () => Throttle.done sortThrottle)))

       val context = if isSome context then SOME
	 (fn (path, p) => (print "sending to context\n";
			   CML.send (valOf context, 
				    Context.ADDEVENT(path,
						     Multicast.recvEvt p))))
		     else NONE

       (* ok, we now can do the search ! *)
     in
       print "starting search...\n";

       (* would be nice to spawn a thread here, but then we can't catch
	AbortSearch below; we should think about this *)
       Dataset.search user dsetpath depth search context hardThrottle doinh;
       {comm=comm, contexts=contexts, user=SOME user,
	saslconn=saslconn} before
       print "done search\n"
     end
       handle Dataset.NoSuchDataset => (M.send (comm, R.NO(SOME tag, 
					[R.NOEXIST ("\"" ^ dset ^ "\"")],
					"no such dataset"));
					conn)
	    | Dataset.DatasetRefer l => (M.send (comm,
					       R.NO(SOME tag,
			      [R.REFERCODE l], "see other server"));
				       conn)
	    | BadSearchCriteria => (M.send (comm, R.BAD(SOME tag,
					       	"bad search criteria")); conn)
	    | BadSearchModifier => (M.send (comm, R.BAD(SOME tag,
						"bad search modifier")); conn)
	    | Comparator.NoSuchComparator s => 
	                           (M.send (comm, R.BAD(SOME tag,
				       "no such comparator " ^ s)); conn)
	    | BadAttributePattern => (M.send (comm, R.BAD(SOME tag,
			               "bad attribute pattern")); conn) 
	    | BadAttributeMetadata => (M.send (comm, R.BAD(SOME tag,
				       "no such metadata")); conn)
	    | AbortSearch => conn)
    | searchdset _ = raise (InternalError "unauthed user calling searchdset")

  (* there's a large amount of duplication between here and above *)
  fun searchcontext(conn as {comm, contexts, user=(SOME user),...}:conn,
		    tag, cname, sm, sc) =
    (let
       (* find the context *)
       val (cname, enum, notify, ch) = 
	 (case (List.find (fn (n, _, _, _) => n = cname) contexts) of
	    NONE => raise NoSuchContext
	  | (SOME x) => x)

       (* extract options *)
       fun get nil (opts as (depth, context, hard, soft, ret, sort, doinh)) = 
	             opts
	 | get ((A.Depth x)::l) (NONE,c,h,s,r,t,i) = 
	             raise BadSearchModifier
	 | get ((A.HardLimit x)::l) (d,c,NONE,s,r,t,i) = 
		     get l (d,c,SOME x,s,r,t,i)
	 | get ((A.Limit x)::l) (d,c,h,NONE,r,t,i) = 
		     get l (d,c,h,SOME x,r,t,i)
	 | get ((A.MakeContext x)::l) (d,NONE,h,s,r,t,i) = 
	             raise BadSearchModifier (* XXX? *)
	 | get ((A.Return x)::l) (d,c,h,s,NONE,t,i) = 
		     get l (d,c,h,s,SOME x,t,i)
	 | get ((A.Sort x)::l) (d,c,h,s,r,NONE,i) = 
		     get l (d,c,h,s,r,SOME x,i)
	 | get ((A.Noinherit)::l) (d,c,h,s,r,t,NONE) =
		     get l (d,c,h,s,r,t,SOME false)
	 | get _ _ = raise BadSearchModifier

       val (_, _, hard, soft, ret, sort, doinh) = 
	 get sm (NONE, NONE, NONE, NONE, NONE, NONE, NONE)

       val doinh = getOpt (doinh, true)

       (* compile search criteria *)
       val (search, time) = compileCSearch (user, doinh) enum sc

       (* compile sort criteria *)
       val sort = Option.map (compileSort (user, doinh)) sort

       (* compile the return attributes; yields a function that sends to
	the client *)
       val metaret = Option.map (compileMetaData (user, doinh)) ret

       val ret = compileRet (comm, tag) (fn (p, e) => p ^ "/" ^ 
					 (Dataset.nameEntry (user, e))) metaret

       val retmodtime = R.TAGMODTIME(tag,
				     (AcapTime.toString o AcapTime.now) ())

       (* compile the soft limit *)
       val softThrottle = 
	 case soft of
	   NONE => Throttle.createpass ret (donePass (tag, comm))
	 | (SOME(a,b)) => Throttle.createsoft (a,b) ret (doneSoft (tag, comm))

       val sortThrottle =
	 (case sort of
	    NONE => softThrottle
	  | (SOME f)  => 
	      Throttle.createsort f (Throttle.call softThrottle)
	                          (fn OK => Throttle.done softThrottle))

       fun doneHard f (Throttle.H_OK) = f () (* call the done on the next fn *)
	 | doneHard f (Throttle.H_WAYTOOMANY) =
	 let
	   (* ok, this should kill the search in progress & report failure *)

	   (* send NO *)
	   val _ = M.send (comm, R.NO(SOME tag, [R.WAYTOOMANY] ,
				      "hard limit exceeded"))
	 in
	   (* finally, tell the search to stop *)
	   raise AbortSearch
	 end

       val hardThrottle =
	 (case hard of
	    NONE => sortThrottle (* skip it *)
	  | (SOME x) => Throttle.createhard x (Throttle.call sortThrottle)
	                  (doneHard (fn () => Throttle.done sortThrottle)))
	    
       fun cancelTime t =
	 M.send (comm, R.NO(SOME tag, [R.MODIFIED (AcapTime.toString t)],
			    "context has been modified"))
     in
       CML.send (ch, Context.SEARCH(search, hardThrottle, 
				    (time, cancelTime))); 
       conn
     end
       handle NoSuchContext => (M.send (comm, R.NO(SOME tag, nil,
						   "no such context"));
	                        conn)
	    | BadSearchCriteria => (M.send (comm, R.BAD(SOME tag,
					       	"bad search criteria")); conn)
	    | BadSearchModifier => (M.send (comm, R.BAD(SOME tag,
						"bad search modifier")); conn)
	    | Comparator.NoSuchComparator s =>
	                           (M.send (comm, R.BAD(SOME tag,
				       "no such comparator " ^ s)); conn) 
	    | AcapTime.BadTimeString => (M.send (comm, R.BAD(SOME tag,
				       "bad time string")); conn)
	    | BadAttributePattern => (M.send (comm, R.BAD(SOME tag,
			               "bad attribute pattern")); conn) 
	    | BadAttributeMetadata => (M.send (comm, R.BAD(SOME tag,
				       "no such metadata")); conn))
    | searchcontext _ = 
       raise (InternalError "unauthed user calling searchcontext")

  (* stores are suppose to be atomic across the whole command.
   that sucks, so we're atomic across entries only *)
  fun storedata (conn as {comm,...} : conn) tag nil = 
      (M.send (comm, R.OK(SOME tag, nil, "store completed"));
       conn)
    | storedata (conn as {comm,user=SOME user,...}) tag 
      ((A.StoreEntry(e, sm, attr))::rest) =
    (let
       val dsetpath = splitDSet user e

       fun splitlast (nil, _) = raise BadEntryPath
	 | splitlast ([s], l) = (s, List.rev l)
	 | splitlast ((a::s), l) = splitlast (s, (a::l))

       (* last component is entry name *)
       val (entry, dsetpath) = splitlast (dsetpath, nil)

       (* extract options *)
       fun get nil (opts as (nocreate, usince)) = opts
	 | get ((A.NoCreate)::l) (NONE, usince) = get l (SOME true, usince)
	 | get ((A.UnchangedSince t)::l) (n,NONE) = get l (n, SOME t)
	 | get _ _ = raise BadStoreModifier

       val (nocreate, unchangedsince) = get sm (NONE, NONE)

       val nocreate = getOpt (nocreate, false)
       val unchangedsince = Option.map AcapTime.fromString unchangedsince

       fun convattr nil (acl, value) = (acl, value)
	 | convattr (("acl",av)::l) (NONE,v) = 
	           convattr l (SOME (Acl.fromValue av), v)
	 | convattr (("value",av)::l) (a,NONE) = 
		   convattr l (a, SOME av)
	 | convattr _ _ = raise BadStoreMetadata

       fun makestore (A.AttrStoreValue(a, v)) = {acl=NONE, attribute=a,
					       value=(SOME v)}
	 | makestore (A.AttrStoreMeta(a, l)) =
	 let
	   val (acl, value) = convattr l (NONE, NONE)
	 in
	   {acl=acl, attribute=a, value=value} : Dataset.storedata
	 end

       (* we pass the full name ! *)
       val _ = Dataset.store user dsetpath (not nocreate) 
	                     ((entry,(List.map makestore attr))::nil)
	                     unchangedsince
    in
      storedata conn tag rest;
      conn
    end
      handle (Dataset.Modified t) => (M.send (comm, R.NO(SOME tag,
		[R.MODIFIED (AcapTime.toString t)], "entry modified"));
				    conn)
           | (Dataset.EnforcedValidation (e,a)) => 
	(M.send (comm, R.NO(SOME tag,
		[R.INVALID (e,a)], "enforced validation"));
					      conn)
           | (Dataset.SyntaxError msg) => (M.send (comm, R.BAD(SOME tag,
							       msg));
				       conn)
	   | (Dataset.DatasetRefer l) => (M.send (comm, R.NO(SOME tag,
				  [R.REFERCODE l], "see other dataset"));
					  conn)
	   | (Dataset.RenameEntry n) => (M.send (comm, R.NO(SOME tag, nil,
				     ("try renaming entry " ^ n ^ " first")));
					 conn)
	   | (Dataset.NotPermitted) => (M.send (comm, R.NO(SOME tag, 
				     [R.PERMISSION "\"\""], "no permission"));
					conn)
	   | Dataset.NoSuchDataset => (M.send (comm, R.NO(SOME tag, nil,
				     ("no such dataset")));
				       conn)
	   | BadEntryPath => (M.send (comm, R.BAD(SOME tag, "bad entry path"));
			      conn)
	   | BadStoreModifier => (M.send (comm, R.BAD(SOME tag,
						      "bad store modifier"));
				  conn)
	   | BadStoreMetadata => (M.send (comm, R.BAD(SOME tag,
						      "bad store metadata"));
				  conn)
	   | (AcapTime.BadTimeString) => (M.send (comm, R.BAD(SOME tag,
							  "bad time string"));
					  conn))
    | storedata _ _ _ = raise (InternalError "unauthed user calling storedata")

  fun run (conn as ({comm,...} : conn)) (A.Error s) =
    (M.send (comm, R.BAD(NONE, s));
     conn)
    | run (conn as ({comm,...})) (A.BareString _) =
    (M.send (comm, R.BAD(NONE, "bare string not expected"));
     conn)
    | run (conn as ({comm,...} : conn)) (A.Line(tag, A.Noop)) =
    (M.send (comm, R.OK(SOME tag, nil, "noop noop noop"));
     conn)
    | run (conn as {comm,...}) (A.Line(tag, A.Logout)) =
    (M.send (comm, R.BYE("have a nice day"));
     M.send (comm, R.OK(SOME tag, nil, "LOGOUT completed"));
     raise CloseConnection)
    | run (conn as {comm,...}) (A.Line(tag, A.Lang(_))) =
    (M.send (comm, R.BAD(SOME tag, "you're stuck with what you have"));
     conn)
    | run ({comm,contexts,saslconn,user=NONE}) 
          (A.Line(tag, A.Authenticate (s,s2))) =
	  (raise doAuth (saslconn, tag, s, s2, comm))
    | run (conn as {comm,user=NONE,...}) (A.Line(tag, _)) =
    (M.send (comm, R.BAD(SOME tag, "authenticate first"));
     conn)
    | run (conn as {comm,...}) (A.Line(tag, A.Authenticate(_,_))) =
    (M.send (comm, R.BAD(SOME tag, "already authenticated"));
     conn)
    | run (conn as {comm,contexts,user,saslconn}) 
          (A.Line(tag, A.FreeContext(c))) =
    (let
       val contexts = deleteContext contexts c
     in
       M.send (comm, R.OK(SOME tag, nil, "context nuked"));
       {comm=comm, contexts=contexts, user=user, saslconn=saslconn}
     end handle
       NoSuchContext => (M.send (comm, R.BAD(SOME tag, "no such context"));
			 conn))
    | run (conn as {comm,contexts,...}:conn) 
          (A.Line(tag, A.UpdateContext(cl))) =
    (let
       val cl = List.map 
	 (fn c => case List.find (fn (n, _, b, _)=>b andalso n=c) contexts of
	  (SOME c) => c
	| NONE => raise NoSuchContext) cl
     in
       (List.app (fn (_, _, _, chan) => 
		  (CML.send (chan, Context.UPDATECONTEXT))) cl;
	M.send (comm, R.OK(SOME tag, nil, "updating contexts"));
	conn)
     end handle
       NoSuchContext => (M.send (comm, R.BAD(SOME tag, "no such context"));
			 conn))
    | run (conn) (A.Line(tag, A.Search(n, sm, sc))) =
      (* figure out if it's a context search or a dataset search *)
    if (String.sub(n, 0) = #"/") then searchdset(conn, tag, n, sm, sc)
    else searchcontext(conn, tag, n, sm, sc)
    | run (conn) (A.Line(tag, A.Store(sel))) =
      storedata conn tag sel
    | run (conn as {user=SOME user,comm,...}) 
          (A.Line(tag, A.DeleteAcl(delobj, ident))) =
      (let
	 val (A.AclObj(dset, attr, ent)) = delobj
	 val aclobj = {dataset=splitDSet user dset,
		       attribute=attr,
		       entry=ent}

	 val _ = Dataset.modifyacl user 
	   (aclobj, Dataset.DELACL(ident))
       in
	M.send (comm, R.OK(SOME tag, nil, "acl deleted"))
       end
	 handle Dataset.NotPermitted =>
	   M.send (comm, R.NO(SOME tag, nil, "not authorized"))
	      | Dataset.NoSuchDataset => (* R.NOEXIST! *)
	   M.send (comm, R.BAD(SOME tag, "no such dataset"));
       conn)
    | run (conn as {user=SOME user,comm,...}) 
          (A.Line(tag, A.ListRights(aobj, auser))) =
      (let 
	 val (A.AclObj(dset, attr, ent)) = aobj
	 val aclobj = {dataset=splitDSet user dset,
		       attribute=attr,
		       entry=ent}

	 val acl = Dataset.modifyacl user (aclobj, Dataset.GETACL)
	 val auser = Acl.Auth.fromString auser

	 val rights = if Acl.Rights.has (Acl.myrights auser acl, 
					 Acl.Rights.ACLa) 
			then ["r", "w", "x", "a", "i"]
		      else nil
       in
	 M.send (comm, R.LISTRIGHTS(tag, "", rights));
	 M.send (comm, R.OK(SOME tag, nil, "Listrights completed"))
       end 
	 handle Dataset.NotPermitted => 
	   M.send (comm, R.NO(SOME tag, nil, "not authorized"))
	      | Dataset.NoSuchDataset => (* R.NOEXIST! *)
	   M.send (comm, R.BAD(SOME tag, "no such dataset"));
	conn)
    | run (conn as {comm,user=SOME user,...}) 
          (A.Line(tag, A.MyRights(aobj))) =
      (let 
	 val (A.AclObj(dset, attr, ent)) = aobj
	 val aclobj = {dataset=splitDSet user dset,
		       attribute=attr,
		       entry=ent}

	 val acl = Dataset.modifyacl user (aclobj, Dataset.GETACL)
	 val rights = Acl.myrights user acl
       in
	 M.send (comm, R.MYRIGHTS(tag, Acl.Rights.toString rights));
	 M.send (comm, R.OK(SOME tag, nil, "Myrights completed"))
       end
	 handle Dataset.NotPermitted =>
	   M.send (comm, R.NO(SOME tag, nil, "not authorized"))
	      | Dataset.NoSuchDataset => 
	   M.send (comm, R.BAD(SOME tag, "no such dataset"));
       conn)
    | run (conn as {comm,user=SOME user,...}) 
          (A.Line(tag, A.SetAcl(aobj, aid, arght))) =
      (let
	 val (A.AclObj(dset, attr, ent)) = aobj
	 val aclobj = {dataset=splitDSet user dset,
		       attribute=attr,
		       entry=ent}

	 val rights = Acl.Rights.fromString arght

	 val acl = Dataset.modifyacl user (aclobj, Dataset.SETACL(aid, rights))
       in
	 M.send (comm, R.OK(SOME tag, nil, "Setacl completed"))
       end
	 handle Acl.Rights.InvalidRights =>
	   M.send (comm, R.BAD(SOME tag, "bad rights"))
	      | Dataset.NotPermitted => 
	   M.send (comm, R.NO(SOME tag, nil, "not authorized"))
	      | Dataset.NoSuchDataset => 
	   M.send (comm, R.BAD(SOME tag, "no such dataset"));
       conn)
    | run (conn as {comm,...}:conn) (A.Line(tag, _)) =
      (M.send (comm, R.NO(SOME tag, nil, "not implemented"));
       conn)

  (* should i attempt to kill all the contexts i'm holding? *)
  fun error (conn as ({comm,...} : conn)) msg =
    (M.send (comm, R.ALERT(msg));
     M.send (comm, R.BYE("serious internal error"));
     conn)

  fun stop (conn as ({comm,contexts,...} : conn)) =
    (app (ignore o (deleteContext contexts) o #1) contexts)

end
